% FAMILY   Compute smoother at several scales and plot them
%
% Parameters:
%    x      x-coordinates of the data
%    y      y-coordinates of the data
%    opt    optional: string of options (case-insensitive), see below
%    alpha  optional, not used, for compatibility only
%    varhat optional, not used, for compatibility only
%    fsp    optional: family smoothing parameters
%
% Options are:
%    'S'    spline smoother (SS), default
%    'L'    local linear smoother (LLR)
%
% fsp is a vector of smoothing parameters (lambda for SS, h for LR).
% If not specified, a suitable set is automatically selected.
%
% Returns:
%    fsp    the set of smoothing parameters used
%    xx     interpolation points
%    ggg    all the smoothers at xx points

function [fsp,xx,ggg]=family(x,y,opt,alpha,varhat,fsp)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Validate parameters and initialize things

L = max(x)-min(x);			% width of design
f = 1/L;				% average density
n = length(x);				% number of design points
nn = 300;				% number of interpolation points
xx = linspace(min(x),max(x),nn)';	% grid of interpolation points

msg = nargchk(2,inf,nargin);
if not(isempty(msg)),
   error(msg);
end

if any(size(x) ~= [n 1]),
   error('x should be column vector');
end

if any(size(x) ~= size(y)),
   error('y should be same size as x');
end

if nargin<3,
   opt = '';
end

S_SPLINE = 1;
S_LLR = 2;
smoother = S_SPLINE;			% default smoother
opt = upper(opt);
if findstr(opt,'S'),
   smoother = S_SPLINE;
end
if findstr(opt,'L'),
   smoother = S_LLR;
end
smoothernames = {'Spline','LLR'};

if nargin<6,
   % automatic selection of smoothing parameters
   fsp = fspselect(x,smoother);
end

linecolors = 'rygcbbmm';
linewidths = linspace(3,1,length(fsp));

disp(sprintf('--- %s family plot ---', smoothernames{smoother}));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute and plot the splines

cla;
hold on;
title(sprintf('%s family plot',smoothernames{smoother}));

ggg = [];

for isp=1:length(fsp),
   sp = fsp(isp);
   if smoother==S_SPLINE,
      [g,u,v] = myspline(x,y,sp);
      gg = mysplineinter(x,g,u,v,xx);
   else
      gg = llr(xx,x,y,sp);
   end
   
   ggg = [ggg gg];			% append to list of smoothers
   
   plot(xx,gg,'-',...
	'linewidth',linewidths(isp),...
	'color',linecolors(isp));
end

plot(x,y,'k.','markersize',8);		% data points

set(gca,'xlim',[xx(1) xx(end)]);
grid on;
drawnow

