% llr.m -- Locar linear (LLR) regression estimator
%
% Parameters:
%		xx		point (or column vector of points) where the estimator is to be evaluated
%		x		column vector of sample points
%		y		column vector of the function values at the sample points;
%           note: may also be a matrix of several columns
%     h     smoothing parameter
%		K		(optional) one-dimensional kernel function, defaults to 'gauss';
%           note: K must return kernel derivative as its second output.
%
% Returns:
%		g		estimated function value at the x points (as a column vector)
%     u     estimated derivative value at the x points (as a column vector)
%     b     slope of the local fit at the x points (generally not same as u)
function [g,u,b]=llr(xx,x,y,h,K)

if nargin<5
   K='gauss';
end

xx = xx(:);						% make sure xx is a column vector

flagu = (nargout >= 2);		% need to evaluate true derivative?
flagb = (nargout >= 3);		% need to evaluate local slope?

N = length(xx);				% number of evaluation points
n = length(x);					% number of sample points
M = size(y,2);					% number of data vectors

if size(y,1)~=n,
   error('x and y must have same number of rows');
end
   
nspecial = 0;
g = zeros(N,M);
b = zeros(N,M);
u = zeros(N,M);

for j=[1:N]
   t = xx(j);
   
   if flagu,
      [KhX,DKhX] = DKh(t,x,K,h);
   else
      KhX = Kh(t,x,K,h);
   end
   
   [sKhX,iKhX] = sort(KhX);
   imax1 = iKhX(end);
   imax2 = iKhX(end-1);
   imax3 = iKhX(end-2);
   if (KhX(imax3) < 1e-6 * KhX(imax2)) ...
	  | (KhX(imax2) < 10*eps*KhX(imax1)),
      % Special case: The two largest weights effectively dominate
      % (or the largest one dominates overwhelmingly).
      % To avoid numerical instability, treat this as a simple
      % linear fit between those two sample points.
      nspecial = nspecial+1;
      gweight = zeros(n,1);
      bweight = zeros(n,1);
      uweight = zeros(n,1);
      x1 = x(imax1);
      x2 = x(imax2);
      
      gweight(imax1) = (x2-t)/(x2-x1);
      gweight(imax2) = (t-x1)/(x2-x1);
      g(j,:) = gweight' * y;
      
      if flagb,
	 bweight(imax1) = 1/(x1-x2);
	 bweight(imax2) = 1/(x2-x1);
	 b(j,:) = bweight' * y;
      end
      
      if flagu,
	 uweight(imax1) = 1/(x1-x2);
	 uweight(imax2) = 1/(x2-x1);
	 u(j,:) = uweight' * y;
      end
   else
      % Regular case
      s0 = sum(KhX);
      s1 = sum((t-x) .* KhX);
      s2 = sum((t-x).^2 .* KhX);
      den = s2*s0-s1^2;					% denominator
      
      W = s2-s1*(t-x);
      gweight = W .* KhX / den;
      g(j,:) = gweight' * y;
      
      if flagb,
	 bweight = (s1-s0*(t-x)) .* KhX / den;
	 b(j,:) = bweight' * y;
      end
      
      if flagu,
	 Ds0 = sum(DKhX);
	 Ds1 = sum((t-x) .* DKhX) + s0;
	 Ds2 = sum((t-x).^2 .* DKhX) + 2*s1;
	 
	 Dden = Ds2*s0 + s2*Ds0 - 2*s1*Ds1;
	 Dinvden = -Dden / den^2;
	 DW = Ds2 - s1 - Ds1*(t-x);
	 
	 uweight = (DW .* KhX / den) ...
		   + (W .* DKhX / den) ...
		   + (W .* KhX * Dinvden);
	 u(j,:) = uweight' * y;
      end
   end
end;

if nspecial>0,
%   warning(sprintf(['llr done as interpolation to avoid numerical' ...
%		    ' instability, in %d points out of %d'], nspecial, N));
end

function ker=Kh(xx,x,K,h)
% Evaluate scaled kernel function at given points
ker=feval(K, (xx-x)/h) / h;


function [ker,der]=DKh(xx,x,K,h)
% Evaluate scaled kernel function and its derivative at given points
[ker,der]=feval(K, (xx-x)/h);
ker = ker/h;
der = der/h^2;


