% MYSPLINEINTER   Interpolate (or extrapolate) a spline at given points
%
% Parameters:
%    X   (sorted) knot points of the spline
%    g   spline value at knots
%    u	first derivative at knots
%    v   second derivative at knots
%    XX  (sorted) evaluation points
%
% Returns:
%    gg  spline value at XX
%    uu  first derivative at XX
%    vv  second derivative at XX
%
% Note: g,u,v can be matrices (of same size). In that case each column
% is taken as a different spline, and interpolated.
%
% See also: MYSPLINE, MYSPLINEKERNEL
%
% Copyright (C) 2003-2004 Jukka Kohonen

function [gg,uu,vv]=mysplineinter(X,g,u,v,XX)

n = length(X);		% number of knot points
N = size(g,2);		% number of splines
m = length(XX);	% number of interpolating points
X = [X; inf];		% add sentinel

% make sure inputs are compatible
if size(X) ~= [n 1], error 'X should be column vector'; end
if size(g) ~= [n N], error 'g should have as many rows as X'; end
if size(u) ~= [n N], error 'u should be same size as g'; end
if size(v) ~= [n N], error 'v should be same size as g'; end
if size(XX) ~= [m 1], error 'XX should be column vector'; end

gg = zeros(m,N);
uu = zeros(m,N);
vv = zeros(m,N);

% Start from leftmost bracket (from -inf to first knot)
left = 0;
right = 1;

for i=1:m,
   x=XX(i);
   
   % move forward until in correct bracket for this point
   while x > X(right),
      left = left+1;
      right = right+1;
   end
   
   if left==0,
      % linear extrapolation leftwards from first knot
      vv(i,:) = 0;
      uu(i,:) = u(1,:);
      gg(i,:) = g(1,:) - (X(1)-x) * u(1,:);
   elseif right==n+1,
      % linear extrapolation rightwards from last knot
      vv(i) = 0;
      uu(i,:) = u(n,:);
      gg(i,:) = g(n,:) + (x-X(n)) * u(n,:);
   else
      % cubic interpolation between two knots
      h = X(right)-X(left);	% distance between knots
      dl = x-X(left);	% distance to left knot
      dr = X(right)-x;	% distance to right knot
      vv(i,:) = (dl*v(right,:) + dr*v(left,:)) / h;
      uu(i,:) = u(left,:) + dl*(v(left,:)+vv(i,:))/2;
      gg(i,:) = (dl*g(right,:) + dr*g(left,:)) / h ...
         - (1/6)*dl*dr * ((1+dl/h)*v(right,:) + (1+dr/h)*v(left,:));
   end
end
