function [G,U,V] = mysplinekernel(X,lambda)
% MYSPLINEKERNEL   Construct kernels for natural cubic smoothing spline.
%
% [G,U,V] = MYSPLINEKERNEL(X,alpha) constructs kernel (or "hat") matrices G,U,V
% for data points located at X, and smoothing parameter alpha.
%
% The matrices can then be used to compute splines for arbitrary data Y
% as follows:
%   g = G*Y   gives the spline value at knots (X points)
%   u = U*Y   gives first derivative of spline
%   v = V*Y   gives second derivative of spline
%
% Note: The kernel computation is relatively slow, as it involves
% two matrix inversions. For computing just one smoothing spline
% for given data Y, MYSPLINE is faster.
%
% Copyright (C) 2003 Jukka Kohonen
%
% Algorithm adapted from Green & Silverman: "Nonparametric Regression
% and Generalized Linear Models: A Roughness Penalty Approach".

n = length(X);
h = diff(X);

% make sure X is a COLUMN vector
if size(X) ~= [n 1], error 'X should be column vector'; end

% make sure X is strictly increasing
if any(diff(X) <= 0), error 'X points must be strictly increasing'; end

t1 = cputime;

% compute R and Q for constructing spline and second derivative
Rsuper  = (1/6) *  h(1:n-2);
Rmain   = (1/3) * (h(1:n-2) + h(2:n-1));
Rsub    = (1/6) *  h(2:n-1);
R = spdiags([Rsub Rmain Rsuper], -1:1, n-2, n-2);
Qsuper  =   1 ./ h(1:n-2);
Qmain   = - 1 ./ h(1:n-2) - 1 ./ h(2:n-1);
Qsub    =   1 ./ h(2:n-1);
Q = spdiags([Qsub Qmain Qsuper], -2:0, n, n-2);

% compute A and B for constructing first derivative
Asuper = [0; 1/h(1); zeros(n-2,1)];
Amain  = [-1/h(1); 1./h];
Asub   = [-1./h; 0];
A = spdiags([Asub Amain Asuper], -1:1, n, n);
Bsuper = [0; -h(1)/6; zeros(n-2,1)];
Bmain  = [-h(1)/3; h/3];
Bsub   = [h/6; 0];
B = spdiags([Bsub Bmain Bsuper], -1:1, n, n);

% Then the hard work:
RaQ = R + lambda*Q'*Q;
V = RaQ \ full(Q');
G = eye(n) - lambda*Q*V;
V = [zeros(1,n); V; zeros(1,n)];
U = A*G + B*V;

t2 = cputime;
%disp(sprintf('spline kernels done in %g', t2-t1));
