% SIZER   Compute and plot SiZer map to indicate significance of trends
%
% Parameters:
%    x        x-coordinates of the data
%    y        y-coordinates of the data
%    opt      optional: string of options (case-insensitive), see below
%    alpha    optional: significance level (default=0.05)
%    varhat   optional: estimate(s) of data variance
%    fsp      optional: family smoothing parameters
%
% Options can be used to specify:
%    * smoother method:   'S'=spline (default), 'L'=local linear
%    * conf.int. method:  'N'=normal, 'B'=bootstrap (default)
%    * error est. method: 'G'=GCV smoother (default), 'V'=Marron's
%                         varying smoother, '2'=GSJ 2nd differences
%    * simultaneity:      'P'=pointwise, 'X'=horizontal (default),
%                         'F'=full simultaneity
%
% varhat can be
%    * scalar: homoscedastic, independent errors
%    * vector: heteroscedastic, independent errors
% Note that specifying varhat overrides any other error estimation
% method ('G','V','2'), and enforces multinormality assumption.
%
% fsp specifies the *family* smoothing parameters (SS: lambda, LLR: h).
% It is used only for marking the smoothing levels in the map.
% If not specified, a suitable set is automatically selected.
%
% Returns:
%    msp       the set of smoothing parameters used in the map
%    xx        interpolation points
%    alluu     all the trends at xx points (N*M matrix,
%              for N interpolation points and M smoothing levels)
%    alldelta  all the CI widths (N*M matrix)
%    allq      all the CI width factors (size depends on simultaneity)
%    map       the SiZer map

function [msp,xx,alluu,alldelta,allq,map]=sizer(x,y,opt,alpha,varhat,fsp)

C0 = cputime;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Validate parameters and initialize things
%
% Note for reader: This section is pretty long and boring.
% Real work begins in next section, over 100 lines down from here.

msg = nargchk(2,6,nargin);
if not(isempty(msg)),
   error(msg);
end

L = max(x)-min(x);			% width of design
f = 1/L;							% average density
n = length(x);					% number of design points
N = 300;						% number of interpolation points
xx = linspace(min(x),max(x),N)';	% grid of interpolation points
B  = 600;						% number of bootstrap samples (if bootstrap is used)

if any(size(x) ~= [n 1]),
   error('x should be column vector');
end

if any(size(x) ~= size(y)),
   error('y should be same size as x');
end

if nargin<3,
   opt = '';
end

opt = upper(opt);

S_SPLINE = 1;
S_LLR = 2;
smoothernames = {'Spline','LLR'};	% displayable names of smoothers
spnames = {'lambda','h'};				% displayable names of smoothing parameters
smoother = S_SPLINE;			% default smoother
if findstr(opt,'S'),
   smoother = S_SPLINE;
end
if findstr(opt,'L'),
   smoother = S_LLR;
end

SIM_PTWISE = 1;
SIM_HORIZ  = 2;
SIM_FULL   = 3;
simnames = {'Pointwise','Simultaneous','Fully simultaneous'};
sim = SIM_HORIZ;				% default simultaneity
if findstr(opt,'P'),
   sim = SIM_PTWISE;
end
if findstr(opt,'X'),
   sim = SIM_HORIZ;
end
if findstr(opt,'F'),
   sim = SIM_FULL;
end

C_NORMAL = 1;
C_BOOT = 2;
cinames = {'normal','bootstrap'};
cimethod = C_BOOT;			% default CI method
if findstr(opt,'N'),
   cimethod = C_NORMAL;
end
if findstr(opt,'B'),
   cimethod = C_BOOT;
end

if cimethod==C_NORMAL & sim==SIM_FULL,
   error('Sorry, fully simultaneous gaussian CI''s not supported');
end

E_GCV = 1;
E_VARYING = 2;
E_2NDDIFF = 3;
E_GIVEN = 4;
enames = {'GCV residuals','varying residuals',...
	  'GSJ 2nd diff','hom. variance as input',...
	  'het. variance as input'};
emethod = E_GCV;				% default error estimation method
if findstr(opt,'G'),
   emethod = E_GCV;
end
if findstr(opt,'V'),
   emethod = E_VARYING;
end
if findstr(opt,'2'),
   emethod = E_2NDDIFF;
end

if nargin<4 | isempty(alpha),
   alpha = 0.05;				% default significance level
end
if alpha > 1.0,
   error('invalid significance level: must be from 0 to 1');
end
if alpha > 0.5,
   disp('WARNING: Suspicious significance level, should probably be <0.5 ?');
end

if nargin>=5 & not(isempty(varhat)),
   emethod = E_GIVEN;
   
   % validate the variance estimate
   varsize = size(varhat);
   if all(varsize == [1 1]),
      vardim = 0;
      Sigma = speye(n) * varhat;
   elseif all(varsize == [n 1]),
      vardim = 1;
      Sigma = spdiags(varhat,0,n,n);
   else
      error('varhat must be scalar or n*1 vector');
   end
   if any(eig(Sigma) < 0),
      error('variances must be nonnegative');
   end
   % Sigma is a diagonal matrix: Elementwise square root is enough,
   % since it equals the matrix square root.
   SqrtSigma = sqrt(Sigma);
end

foobar = emethod;
if foobar==E_GIVEN & vardim==1,
   foobar = foobar+1;
end
emethodname = enames{foobar};

if nargin<6,
   % automatic selection of family smoothing parameters
   % (for compatibility with FAMILY)
   hlist = L * 10.^[0:-0.5:-2];
   if smoother == S_SPLINE,
      lambdalist = hlist.^4 * n * f;	% this will have the specified h
      fsp = lambdalist;
   else
      hlist = hlist*1.25;					% approx. same width as spline
      fsp = hlist;
   end
end
logfsp = log10(fsp);

linecolors = 'rygcbbmm';
linewidths = linspace(3,1,length(fsp));

disp('--- SiZer map ---');
disp(sprintf('  alpha:     %g', alpha));
disp(['  Smoother:  ' smoothernames{smoother}]);
disp(['  CI''s:      ', simnames{sim}, ' ', cinames{cimethod}]);
if emethod==E_GCV,
   disp(['  Errors:    GCV fit residuals']);
elseif emethod==E_VARYING,
   disp(['  Errors:    varying fit residuals (Marron)']);
elseif emethod==E_2NDDIFF,
   disp(['  Errors:    2nd diff (Gasser-Sroka-Jennen-Steinmetz)']);
elseif emethod==E_GIVEN,
   if vardim==0,
      disp(['  Errors:    normal, iid., given variance']);
   elseif vardim==1,
      disp(['  Errors:    normal, independent, given variances']);
   else
      error 'impossible';
   end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialize map dimensions and colors
mapheight = 50;
mapwidth  = length(xx);
logfsp = log10(fsp);
fspacing = abs(mean(diff(logfsp)));

mapleft   = xx(1);
mapright  = xx(end);
maptop    = max(logfsp)+fspacing/2;
mapbottom = min(logfsp)-fspacing/2;

logmsp = linspace(mapbottom,maptop,mapheight);
msp = 10.^logmsp;

colormap(...
   [.35 .35 .35; ...		% gray for no inference (CI unreliable)
      1 0 0; ...			% red for positive (CI fully above zero)
      .5 0 .5; ...		% magenta for indeterminate (CI contains zero)
      0 0 1; ...			% blue for negative (CI fully below zero)
   ]);
MAPGRAY = 1;
MAPPOS = 2;
MAPZERO = 3;
MAPNEG = 4;

map = ones(mapheight,mapwidth)*MAPZERO;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Precompute hat matrices
allG = zeros(n,n,mapheight);
allU = allG;
allV = allG;

for i=1:mapheight,
   sp = msp(i);
   
   if smoother==S_SPLINE,
      [G,U,V] = mysplinekernel(x,sp);
   else
      [G,U] = llrkernel(x,x,sp);
      V = ones(n,n)*nan;	% not needed, not computed
   end
   
   allG(:,:,i) = G;
   allU(:,:,i) = U;
   allV(:,:,i) = V;
end

C1 = cputime;
disp(sprintf('hat matrices done in %.1f cpu seconds', C1-C0));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Select best-fit smoother (using GCV)
gcvscore = zeros(mapheight,1);
for i=1:mapheight,
   G = allG(:,:,i);
   g = G*y;
   % GCV score according to eq. 3.13 of Green&Silverman
   % but modified to ignore cases where trace(G)==n.
   if trace(G)==n,
      gcvscore(i) = NaN;
   else
      gcvscore(i) = (1/n) * sum((y-g).^2) / ((1-trace(G)/n).^2);
   end
end
[dummy,igcv] = min(gcvscore);
spgcv = msp(igcv);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Estimate data errors and variance using various methods

% GCV fit residuals
[gcv_var,gcv_eps] = residuals(allG(:,:,igcv), y);

% 2nd differences (GSJ)
[gsj_var,gsj_eps] = resid_gsj(x,y);

disp('std error estimates:');
disp(sprintf('  GCV fit:   %.4g  [at lg(%s)=%.2f]', ...
   sqrt(gcv_var), spnames{smoother}, log10(spgcv)));
disp(sprintf('  2nd dif:   %.4g', sqrt(gsj_var)));

if emethod==E_GCV,
   varhat = gcv_var;
   epshat = gcv_eps;
   Sigma  = speye(n) * varhat;
elseif emethod==E_2NDDIFF,
   varhat = gsj_var;
   epshat = gsj_eps;
   Sigma  = speye(n) * varhat;
end
drawnow   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct the map

alluu = zeros(N,mapheight);
alldelta = zeros(N,mapheight);
if sim==SIM_HORIZ,
   allq = zeros(1,mapheight);
end   

if sim==SIM_FULL,
   % Special code for fully simultaneous confidence intervals.

   if emethod==E_VARYING,
      error('Sorry, varying error estimates not supported for full simultaneity');
   end
   
   if cimethod~=C_BOOT,
      error('Sorry, full simultaneity is supported only for bootstrap');
   end
   
   % Collect all normalized absolute errors into one huge matrix.
   devlist = zeros(B, mapheight);

   sigmaz = zeros(N, mapheight);

   if emethod~=E_GIVEN
      % We must already have epshat (sample from error distribution).
      epszero = epshat - mean(epshat);	% make sure they are zero-mean
   end

   allUU = zeros(N,n,mapheight);
   
   for i=1:mapheight,
      sp = msp(i);
      h = width(sp,smoother,n,f);
      if smoother==S_SPLINE,
	 G = allG(:,:,i);
	 U = allU(:,:,i);
	 V = allV(:,:,i);
	 [GG,UU] = mysplineinter(x,G,U,V,xx);
      else
	 [GG,UU] = llrkernel(xx,x,sp);
      end
      uu = UU*y;
      alluu(:,i) = uu;
      allUU(:,:,i) = UU;
      
      varz = zeros(N,1);
      for k=1:N,
	 varz(k) = UU(k,:)*Sigma*UU(k,:)';
      end
      sigmaz(:,i) = sqrt(varz);
   end   
   
   for b=1:B,
      if emethod==E_GIVEN,
	 % Generate a parametric bootstrap sample
	 epsstar = SqrtSigma * randn(n,1);
      else
	 % Generate a bootstrap sample from residuals
	 index = ceil(rand(n,1)*length(epszero));
	 epsstar = epszero(index);
      end
         
      % Loop through all smoothing levels and do smoothers
      for i=1:mapheight,
	 UU = allUU(:,:,i);
      
         % Do a smoother over the bootstrapped error sample,
         % to get trend errors
         uerr = UU*epsstar;
         
         % Normalized trend errors
         z = uerr ./ sigmaz(:,i);
         
         % Find largest abs coordinate (at this level, for this sample)
         devlist(b,i) = max(abs(z));
      end
      
      if mod(b,100)==0,
         disp(sprintf('%d/%d bootstraps done',b,B));
      end
   end

   % Find largest abs coordinate of each sample (over all levels)
   fulldevlist = max(devlist,[],2);
   
   % Note: q will be a single scalar for scaling all CI's.
   % Delta will be a matrix of all CI lengths.
   q = prctile(fulldevlist, 100*(1-alpha));
   
   Delta = q * sigmaz;
   alldelta = Delta;
   allq = q;

   disp(sprintf('chose q=%.3g',q));
   for i=1:mapheight,
      pos = (alluu(:,i) > Delta(:,i));
      neg = (alluu(:,i) < -Delta(:,i));
      map(i,find(pos)) = MAPPOS;
      map(i,find(neg)) = MAPNEG;
   end
   
else
   % Pointwise or half-simultaneous intervals.

   for i=1:mapheight,
      sp = msp(i);
      h = width(sp,smoother,n,f);
      
      if smoother==S_SPLINE,
         G = allG(:,:,i);
         U = allU(:,:,i);
         V = allV(:,:,i);
         [GG,UU] = mysplineinter(x,G,U,V,xx);
      else
         [GG,UU] = llrkernel(xx,x,sp);
      end
      
      uu = UU*y;
      alluu(:,i) = uu;
      
      if emethod==E_VARYING,
         % Marron-style error estimation:
         % residuals to this particular smoother
         [varhat,epshat] = residuals(G,y);
         Sigma = speye(n) * varhat;
      end
      
      % First, compute confidence interval lengths...
      if cimethod==C_NORMAL,
         % Gaussian confidence intervals.
         % We could calculate full Sigmaz as follows:
         %   Sigmaz = UU*Sigma*UU';					% covariance of trend errors
         % But we only need diag(Sigmaz), so we do it faster:
         varz = zeros(N,1);
         for k=1:N,
            varz(k) = UU(k,:)*Sigma*UU(k,:)';
         end
         sigmaz = sqrt(varz);
         
         if sim==SIM_PTWISE,
            q       = qnorm(1-alpha/2);
	    allq    = q;
            Delta   = q * sigmaz;			% simple two-sided confidence interval, eq. 5.8
         elseif sim==SIM_HORIZ,
            if smoother==S_SPLINE
               nblocks = L/h;						% heuristic number of blocks, eq. 6.20
            else
               nblocks = 1.25*L/h;				% more heuristic (section 6.2)
            end
            nblocks = max(nblocks,1);
            q       = qnorm((1+(1-alpha)^(1/nblocks))/2);		%
                                                                        % block approximation, eq. 6.15
	    allq(i) = q;
            Delta   = q * sigmaz;			% simultaneous confidence intervals, eq. 6.4
         else
            error('not supported');
         end
      else
         % Bootstrap confidence intervals.
	 
	 if emethod~=E_GIVEN
	    % We must already have epshat (sample from error distribution).
	    epszero = epshat - mean(epshat);	% make sure they are zero-mean
         end
	 
         sigmazfactor = sqrt(sum(UU.^2, 2));
         % corresponds to v_j in H&E, 2001, equation (3.14).
         % Could be calculated as sqrt(diag(UU*ones(n)*UU')), but
         % that would be more expensive.
         
         if sim==SIM_PTWISE,
            devlist = zeros(N,B);		% collect abs of each coordinate (interval!)
         else
            devlist = zeros(1,B);		% collect abs ov MAX coordinate (hypercube!)
         end         
         
         for b=1:B,
	    if emethod==E_GIVEN,
	       % Generate a parametric bootstrap sample
	       epsstar = SqrtSigma * randn(n,1);
	    else
	       % Generate a bootstrap sample from residuals
	       index = ceil(rand(n,1)*length(epszero));
	       epsstar = epszero(index);
	    end
            
            % Do a smoother over the bootstrapped error sample,
            % to get trend errors
            uerr = UU*epsstar;
            
            % Normalized trend errors
            z = uerr ./ sigmazfactor;
            
            if sim==SIM_PTWISE,
               % all abs coordinates (= zero-centered interval lengths)
               devlist(:,b) = abs(z);
            else
               % largest abs coordinate (= hypercube size)
               devlist(1,b) = max(abs(z));
            end
         end
         
         % Note: For pointwise intervals, q will be an n-vector,
         % but for simultaneous ones, q will be scalar.
         q = prctile(devlist', 100*(1-alpha))';
	 if sim==SIM_PTWISE,
	    allq = q;
	 else
	    allq(i) = q;
	 end
         Delta = q .* sigmazfactor;
      end
      
      % ... then see which intervals are fully positive or negative.
      alldelta(:,i) = Delta;
      pos = (uu > Delta);
      neg = (uu < -Delta);
      map(i,find(pos)) = MAPPOS;
      map(i,find(neg)) = MAPNEG;
      
      if mod(i,10)==0,
         disp(sprintf('%d/%d rows done',i,mapheight));
      end
   end
   
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plot the map

cla;
hold on;
title(sprintf('%s %s SiZer map at \\alpha=%g (%s)', ...
	      simnames{sim}, cinames{cimethod}, ...
	      alpha, emethodname));

image([mapleft mapright], [mapbottom maptop], map);
set(gca,'ydir','normal');		% small values down
set(gca,'tickdir','out');

axis([mapleft mapright mapbottom maptop]);		% note: boundary patches will be half size
if smoother==S_SPLINE,
   ylabel('lg \lambda');
else
   ylabel('lg h');
end
set(gca,'ytick',fliplr(logfsp));
set(gca,'yticklabel',num2str(fliplr(logfsp)','%.1f'));


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Extra graphics

% Family smoothing levels
for i=1:length(fsp),
   sp = fsp(i);
   logsp = log10(sp);
   [h,H] = width(sp,smoother,n,f);
   plot([mapleft mapright],[logsp logsp],'w-');
   markeroffset = (mapright-mapleft)/60;
   plot(mapright+markeroffset,logsp,'k<',...
      'markerfacecolor',linecolors(i),'markersize',8,'clip','off');
end

% Effective window widths
xmid = (mapleft+mapright)/2;
mH = zeros(length(msp),1);
for i=1:length(msp),
   sp = msp(i);
   [h,H] = width(sp,smoother,n,f);
   mH(i) = H;
end
plot(xmid-mH, log10(msp), 'w-.');
plot(xmid+mH, log10(msp), 'w-.');

% "Correct" bandwidth, chosen by GCV
plot([mapleft mapright],[log10(spgcv) log10(spgcv)],'w:');


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% That's all!
C2 = cputime;
disp(sprintf('sizer done in %.1f cpu seconds', C2-C0));



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary functions

function [h,H]=width(sp,smoother,n,f)
% convert smoothing parameter to kernel width and window width
if smoother==1,		% S_SPLINE
   h = (sp / n / f) ^ (1/4);
   H = 2.5*h;
else						% S_LLR
   h = sp;
   H = 2*h;
end

function [varhat,epshat] = residuals(G,y)
g = G * y;
n = length(G);
EDD = trace(eye(n) - G);
varhat = sum((y-g).^2) / EDD;
epshat = y-g;
epshat = epshat * sqrt(varhat) / std(epshat);
% todo: something more elegant
