function [g,u,v,fit,rough,R,Q] = myspline(X,Y,lambda,w)
% MYSPLINE   Compute natural cubic smoothing spline at knots.
%
% [g,u,v] = MYSPLINE(X,Y,lambda,w) computes a smoothing spline for
% data points (X,Y), with smoothing parameter lambda,
% and relative weights w (optional; default is equal weights).
%
% X should be a column vector.
% Y may be be a column vector. It can also be a matrix; then splines for each column are made.
%
% g is the value of the spline at knots (X points),
% u is the first derivative of the spline at knots, and
% v is the second derivative.
%
% Copyright (C) 2003-2004 Jukka Kohonen
%
% Algorithm adapted from Green & Silverman: "Nonparametric Regression
% and Generalized Linear Models: A Roughness Penalty Approach".
%
% All matrix operations are done in sparse form, in order to exploit
% the band structure.

n = length(X);
h = diff(X);

weighted = nargin>=4;

% make sure X and Y are compatible
if any(size(X) ~= [n 1]), error 'X should be column vector'; end
if size(Y,1) ~= n, error 'Y should be have as many rows as X'; end
if weighted & any(size(w) ~= [n 1]), error 'w should be column vector and compatible with X'; end

% make sure X is strictly increasing
if any(diff(X) <= 0), error 'X points must be strictly increasing'; end

if weighted,
	Winv = spdiags(1./w, 0, n, n);
end

t1 = cputime;

% compute R and Q for constructing spline and second derivative
Rsuper  = (1/6) *  h(1:n-2);
Rmain   = (1/3) * (h(1:n-2) + h(2:n-1));
Rsub    = (1/6) *  h(2:n-1);
R = spdiags([Rsub Rmain Rsuper], -1:1, n-2, n-2);
Qsuper  =   1 ./ h(1:n-2);
Qmain   = - 1 ./ h(1:n-2) - 1 ./ h(2:n-1);
Qsub    =   1 ./ h(2:n-1);
Q = spdiags([Qsub Qmain Qsuper], -2:0, n, n-2);

if nargout>1,
	% compute A and B for constructing first derivative
	Asuper = [0; 1/h(1); zeros(n-2,1)];
	Amain  = [-1/h(1); 1./h];
	Asub   = [-1./h; 0];
	A = spdiags([Asub Amain Asuper], -1:1, n, n);
	Bsuper = [0; -h(1)/6; zeros(n-2,1)];
	Bmain  = [-h(1)/3; h/3];
	Bsub   = [h/6; 0];
	B = spdiags([Bsub Bmain Bsuper], -1:1, n, n);
end

% Then construct the actual vectors:

% eq. (2.15) for second derivative; or eq. (3.24) for the weighted case
if weighted,
   RaQ = R + lambda*Q'*Winv*Q;
else
   RaQ = R + lambda*Q'*Q;
end
v = RaQ \ (Q'*Y);

% eq. (2.14) or (3.23) for smoother
if weighted,
   g = Y - lambda*Winv*Q*v;
else
   g = Y - lambda*Q*v;
end

% Extra data that could be of interest:
if nargout>3,
	if weighted,
	   fit = sum(w.*(g-Y).^2);
	else
	   fit = sum((g-Y).^2);
	end
   rough = v' * R * v;		% eq. (2.5)
end

% pad v up to same row count as g
% (with zero, since we are constructing a natural spline)
N = size(Y,2);
v = [zeros(1,N); v; zeros(1,N)];

% first derivative
if nargout>1,
   u = A*g + B*v;
end

t2 = cputime;
%disp(sprintf('%d spline(s) done in %g seconds', size(Y,2), t2-t1));
